/*
Module : SocksClient.cpp
Purpose: Implementation for the CSocksClient class
Created: PJN / 16-02-2003
History: PJN / 07-09-2005 1. Updated the code to compile cleanly using the latest version of CWSocket.
                          2. Fixed a typo in the logging of an error if socks 4 clients are not allowed.
         PJN / 15-01-2007 1. Updated copyright details
                          2. Removed CSocksClient destructor as it was not really required.
                          3. Optimized CSocksClient constructor code
                          4. Code now uses newer C++ style casts instead of C style casts.
                          5. General review of code base for correctness.
                          6. Removed unused code from CSocksClient::AllowThisOutboundConnection and
                          CSocksClient::OnEndRelay

Copyright (c) 2003 - 2017 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/

//////////////// Includes ////////////////////////////////////////////

#include "stdafx.h"
#include "SocksClient.h"


//////////////// Macros / Defines ////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//////////////// Implementation //////////////////////////////////////

CSocksClient::CSocksClient() : m_pSettings(nullptr),
                               m_pError(nullptr),
                               m_nRemotePort(0),
                               m_pSocksRequest(nullptr)
{
}

BOOL CSocksClient::InitInstance()
{
  //Allow the thread to be created if the waitable timer was created
  return m_DataTimer.Create(TRUE);
}

bool CSocksClient::AllowThisIncomingConnection()
{
  return true;
}

bool CSocksClient::AllowUser(_In_ const CSocksUsernameAuthenticationDetails& userDetails)
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);

  return (userDetails.sPassword == m_pSettings->m_sPassword) &&
         (userDetails.sUserName == m_pSettings->m_sUsername);
}

bool CSocksClient::SetupInboundDetails()
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);
  ASSERT(m_pError != nullptr);
  ASSERT(m_pSocksRequest != nullptr);

  if (!AllowThisIncomingConnection())
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SetupInboundDetails, Not allowing client connection, client:%s"), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }

  //Use a Win32 event notification on the socket
  int nError = WSAEventSelect(m_pSocksRequest->m_ClientSocket, m_DataEvent1, FD_READ | FD_CLOSE);
  if (nError == SOCKET_ERROR)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SetupInboundDetails, Failed in call to WSAEventSelect on the lan socket, error:%u, client:%s"), GetLastError(), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }

  //Parse out the initial connection request
  if (!m_pSocksRequest->m_ClientSocket.Read(m_ConnectionRequest, m_pSettings->m_dwReadTimeout, m_DataTimer, m_StopEvent, m_DataEvent1))
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SetupInboundDetails, Failed to read connection request from client, client:%s"), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }

  return true;
}

bool CSocksClient::SetupOutboundDetails(_In_ CSocksSocket& destinationSocket, _Out_ bool& bSocketSetupFailure, _Out_ int& nConnectionError)
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);
  ASSERT(m_pError != nullptr);

  ///Now try and make the connection to the remote host
  bool bConnected = false;
  nConnectionError = ERROR_SUCCESS;
  bSocketSetupFailure = false;

  if (m_ConnectionRequest.m_bSocks5Request)
  {
    if (m_RequestDetails.m_Type == CSocks5RequestDetails::IP4)
    {
      m_sRemoteAddress.Format(_T("%d.%d.%d.%d"), static_cast<int>(m_RequestDetails.IP_4.DST_IP.S_un.S_un_b.s_b1), static_cast<int>(m_RequestDetails.IP_4.DST_IP.S_un.S_un_b.s_b2), 
                              static_cast<int>(m_RequestDetails.IP_4.DST_IP.S_un.S_un_b.s_b3), static_cast<int>(m_RequestDetails.IP_4.DST_IP.S_un.S_un_b.s_b4)); 
      m_nRemotePort = m_RequestDetails.IP_4.DSTPORT;
    }
    else if (m_RequestDetails.m_Type == CSocks5RequestDetails::IP6)
    {
      m_sRemoteAddress.Format(_T("%04X.%04X.%04X.%04X.%04X.%04X.%04X.%04X"), static_cast<int>(m_RequestDetails.IP_6.DST_IP6.u.Word[0]), static_cast<int>(m_RequestDetails.IP_6.DST_IP6.u.Word[1]), 
                              static_cast<int>(m_RequestDetails.IP_6.DST_IP6.u.Word[2]), static_cast<int>(m_RequestDetails.IP_6.DST_IP6.u.Word[3]), static_cast<int>(m_RequestDetails.IP_6.DST_IP6.u.Word[4]), 
                              static_cast<int>(m_RequestDetails.IP_6.DST_IP6.u.Word[5]), static_cast<int>(m_RequestDetails.IP_6.DST_IP6.u.Word[6]), static_cast<int>(m_RequestDetails.IP_6.DST_IP6.u.Word[7])); 
      m_nRemotePort = m_RequestDetails.IP_6.DSTPORT;
    }
    else
    {
      ASSERT(m_RequestDetails.m_Type == CSocks5RequestDetails::HostName);
      m_sRemoteAddress = m_RequestDetails.HOSTNAME.DST_HOST.HOST;
      m_nRemotePort = m_RequestDetails.HOSTNAME.DSTPORT;
    }
  }
  else
  {
    m_sRemoteAddress.Format(_T("%d.%d.%d.%d"), static_cast<int>(m_ConnectionRequest.m_Socks4Request.DSTIP.S_un.S_un_b.s_b1), static_cast<int>(m_ConnectionRequest.m_Socks4Request.DSTIP.S_un.S_un_b.s_b2), 
                            static_cast<int>(m_ConnectionRequest.m_Socks4Request.DSTIP.S_un.S_un_b.s_b3), static_cast<int>(m_ConnectionRequest.m_Socks4Request.DSTIP.S_un.S_un_b.s_b4)); 
    m_nRemotePort = m_ConnectionRequest.m_Socks4Request.DSTPORT;
  }

  //Make the connection
  try
  {
  #ifdef CWSOCKET_MFC_EXTENSIONS
    destinationSocket.SetBindAddress(m_pSettings->m_sConnectBindAddress);
  #else
    destinationSocket.SetBindAddress(CWSocket::String(m_pSettings->m_sConnectBindAddress));
  #endif //#ifdef CWSOCKET_MFC_EXTENSIONS
    destinationSocket.CreateAndConnect(m_sRemoteAddress, m_nRemotePort);
    bConnected = true;
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
  catch(CWSocketException* pEx)
  {
    nConnectionError = pEx->m_nError;

    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SetupOutboundDetails, Failed to connect to destination, client:%s, remote host:%s:%d"), m_sLanAddress.GetString(), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    m_pError->OnError(sError);

    pEx->Delete();
  }
#else
  catch(CWSocketException& e)
  {
    nConnectionError = e.m_nError;

    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SetupOutboundDetails, Failed to connect to destination, client:%s, remote host:%s:%d"), m_sLanAddress.GetString(), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    m_pError->OnError(sError);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //Use a Win32 event notification on the socket
  int nError = WSAEventSelect(destinationSocket, m_DataEvent2, FD_READ | FD_CLOSE);
  if (nError == SOCKET_ERROR)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SetupOutboundDetails, Failed in call to WSAEventSelect on the internet socket, error:%u, client:%s"), GetLastError(), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    //Force the failure to be sent back to the client
    bConnected = false; 
    bSocketSetupFailure = true;
  }

  return bConnected;
}

bool CSocksClient::SendConnectionResponsev5(_In_ BYTE byCmd, _In_ SOCKET hDestinationSocket)
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);
  ASSERT(m_pError != nullptr);
  ASSERT(m_pSocksRequest != nullptr);
  ASSERT(m_ConnectionRequest.m_bSocks5Request);

  //Form the response
  SOCKS5_IP4_REQUEST_DETAILS responseIP4;
  SOCKS5_IP6_REQUEST_DETAILS responseIP6;
  BYTE responseHostname[262]; //262 is the maximum size of a hostname response
  BYTE* pbyResponse = nullptr;
  int nResponseSize = 0;
  if (m_RequestDetails.m_Type == CSocks5RequestDetails::IP4)
  {
    pbyResponse = reinterpret_cast<BYTE*>(&responseIP4);
    nResponseSize = sizeof(responseIP4);
    memset(&responseIP4, 0, sizeof(responseIP4));
    responseIP4.Base.VER = 5;
    responseIP4.Base.CMD = byCmd;
    responseIP4.Base.ATYP = 1;
    sockaddr_in sockAddr;
    memset(&sockAddr, 0, sizeof(sockAddr));
    int nSockAddrLen = sizeof(sockAddr);
    if (getsockname(hDestinationSocket, reinterpret_cast<SOCKADDR*>(&sockAddr), &nSockAddrLen) != SOCKET_ERROR)
    {
      responseIP4.DST_IP = sockAddr.sin_addr;
      responseIP4.DSTPORT = htons(sockAddr.sin_port);
    }
  }
  else if (m_RequestDetails.m_Type == CSocks5RequestDetails::IP6)
  {
    pbyResponse = reinterpret_cast<BYTE*>(&responseIP6);
    nResponseSize = sizeof(responseIP6);
    memset(&responseIP6, 0, sizeof(responseIP6));
    responseIP6.Base.VER = 5;
    responseIP6.Base.CMD = byCmd;
    responseIP6.Base.ATYP = 4;
    sockaddr_in6 sockAddr;
    memset(&sockAddr, 0, sizeof(sockAddr));
    int nSockAddrLen = sizeof(sockAddr);
    if (getsockname(hDestinationSocket, reinterpret_cast<SOCKADDR*>(&sockAddr), &nSockAddrLen) != SOCKET_ERROR)
    {
      memcpy_s(&responseIP6.DST_IP6, sizeof(responseIP6.DST_IP6), &sockAddr.sin6_addr, sizeof(sockAddr.sin6_addr));
      responseIP6.DSTPORT = htons(sockAddr.sin6_port);
    }
  }
  else
  {
    ASSERT(m_RequestDetails.m_Type == CSocks5RequestDetails::HostName);
    pbyResponse = responseHostname;
    BYTE* pbyCurrentByte = pbyResponse;
    *pbyCurrentByte = 5;
    ++pbyCurrentByte;
    *pbyCurrentByte = byCmd;
    ++pbyCurrentByte;
    *pbyCurrentByte = 0;
    ++pbyCurrentByte;
    *pbyCurrentByte = 3;
    ++pbyCurrentByte;
    *pbyCurrentByte = m_RequestDetails.HOSTNAME.DST_HOST.LENGTH;
    ++pbyCurrentByte;
    memcpy(pbyCurrentByte, m_RequestDetails.HOSTNAME.DST_HOST.HOST, m_RequestDetails.HOSTNAME.DST_HOST.LENGTH);
    pbyCurrentByte += m_RequestDetails.HOSTNAME.DST_HOST.LENGTH;
    WORD* pPort = reinterpret_cast<WORD*>(pbyCurrentByte);
    *pPort = htons(m_RequestDetails.HOSTNAME.DSTPORT);
    nResponseSize = 7 + m_RequestDetails.HOSTNAME.DST_HOST.LENGTH;
  }

  //Send the response back
  try
  {
    m_pSocksRequest->m_ClientSocket.SendWithRetry(pbyResponse, nResponseSize, m_pSettings->m_dwWriteTimeout);
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
  catch(CWSocketException* pEx)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SendConnectionResponsev5, Failed to send Socks 5 connection response packet, error:%d, client:%s"), pEx->m_nError, m_sLanAddress.GetString());
    m_pError->OnError(sError);

    pEx->Delete();

    return false;
  }
#else
  catch(CWSocketException& e)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SendConnectionResponsev5, Failed to send Socks 5 connection response packet, error:%d, client:%s"), e.m_nError, m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  return true;
}


bool CSocksClient::SendConnectionResponsev4(_In_ bool bConnected)
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);
  ASSERT(m_pError != nullptr);
  ASSERT(m_pSocksRequest != nullptr);
  ASSERT(!m_ConnectionRequest.m_bSocks5Request);

  //Send the reponse back to the client on the connection success or otherwise
  SOCKS4_REPLY_PACKET response;
  memset(&response, 0, sizeof(response));
  response.CD = bConnected ? static_cast<BYTE>(90) : static_cast<BYTE>(91);

  //Send the response back
  try
  {
    m_pSocksRequest->m_ClientSocket.SendWithRetry(&response, sizeof(response), m_pSettings->m_dwWriteTimeout);
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
  catch(CWSocketException* pEx)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SendConnectionResponsev4, Failed to send Socks 4 connection response packet, error:%d, client:%s"), pEx->m_nError, m_sLanAddress.GetString());
    m_pError->OnError(sError);

    pEx->Delete();

    return false;
  }
#else
  catch(CWSocketException& e)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::SendConnectionResponsv4, Failed to send Socks 4 connection response packet, error:%d, client:%s"), e.m_nError, m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  return true;
}


bool CSocksClient::DoSocks5Negotiation()
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);
  ASSERT(m_pError != nullptr);
  ASSERT(m_pSocksRequest != nullptr);

  //Drop the connection if we are not supporting Socks5
  if (!m_pSettings->m_bSupportSocks5)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoSocks5Negotiation, Dropping client connection as we are not configured to support Socks 5 protocol, client:%s"), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }

  //Check to see if any of the clients methods is suitable
  bool bFoundAcceptableMethod = false;
  BYTE nAcceptableMethod = 0;
  for (int i=0; i<m_ConnectionRequest.m_Socks5Request.NMETHODS && !bFoundAcceptableMethod; i++)
  {
    if (m_pSettings->m_bSupportUserNamePassword)
    {
      bFoundAcceptableMethod = (m_ConnectionRequest.m_Socks5Request.METHODS[i] == 2);
      nAcceptableMethod = 2;
    }
    if (!bFoundAcceptableMethod && m_pSettings->m_bSupportAnonymous)
    {
      bFoundAcceptableMethod = (m_ConnectionRequest.m_Socks5Request.METHODS[i] == 0);
      nAcceptableMethod = 0;
    }  
  }

  SOCKS5_METHODSELECTION_MESSAGE methodSelected;
  methodSelected.VER = 5;
  if (bFoundAcceptableMethod)
    methodSelected.METHOD = nAcceptableMethod; 
  else
    methodSelected.METHOD = 0xFF;

  try
  {
    m_pSocksRequest->m_ClientSocket.SendWithRetry(&methodSelected, sizeof(methodSelected), m_pSettings->m_dwWriteTimeout);
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
  catch(CWSocketException* pEx)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoSocks5Negotiation, Failed to send Socks 5 selection method packet, error:%d, client:%s"), pEx->m_nError, m_sLanAddress.GetString());
    m_pError->OnError(sError);

    pEx->Delete();

    return false;
  }
#else
  catch(CWSocketException& e)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoSocks5Negotiation, Failed to send Socks 5 selection method packet, error:%d, client:%s"), e.m_nError, m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //Fail if there were not acceptable connection methods
  if (!bFoundAcceptableMethod)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoSocks5Negotiation, client did not provide any acceptable authentication methods, client:%s"), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }

  //Need to support Username / password Socks 5 authentication method
  if (nAcceptableMethod == 2)
  {
    if (!DoUserNameAuthentication())
      return false;
  }

  //Now read the details from the client
  if (!m_pSocksRequest->m_ClientSocket.Read(m_RequestDetails, m_pSettings->m_dwReadTimeout, m_DataTimer, m_StopEvent, m_DataEvent1))
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoSocks5Negotiation, Failed to read request details from client, client:%s"), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }

  //Check the Command field (Currently we only support CONNECT command)
  bool bCMDAllowed = true;
  switch (m_RequestDetails.m_Type)
  {
    case CSocks5RequestDetails::IP6:
    {
      bCMDAllowed = (m_RequestDetails.IP_6.Base.CMD == 1); 
      break;
    }
    case CSocks5RequestDetails::IP4:
    {
      bCMDAllowed = (m_RequestDetails.IP_4.Base.CMD == 1);
      break;
    }
    case CSocks5RequestDetails::HostName:
    {
      bCMDAllowed = (m_RequestDetails.HOSTNAME.Base.CMD == 1);
      break;
    }
    default:
    {
      bCMDAllowed = false;
      break;
    }
  }

  if (!bCMDAllowed)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoSocks5Negotiation, Sending Socks5 response packet (invalid command type), client:%s"), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    SendConnectionResponsev5(7, INVALID_SOCKET); //Command not supported
    return false;
  }

  return true;
}

bool CSocksClient::DoUserNameAuthentication()
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);
  ASSERT(m_pError != nullptr);
  ASSERT(m_pSocksRequest != nullptr);

  CSocksUsernameAuthenticationDetails userDetails;
  if (!m_pSocksRequest->m_ClientSocket.Read(userDetails, m_pSettings->m_dwReadTimeout, m_DataTimer, m_StopEvent, m_DataEvent1))
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoUserNameAuthentication, Failed to read user authentication details request from client, client:%s"), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }

  //Form the response
  bool bAllowUser = AllowUser(userDetails);
  SOCKS5_USERNAME_AUTHENTICATION_REPLY userResponse;
  userResponse.VER = 1;
  userResponse.STATUS = bAllowUser ? static_cast<BYTE>(0) : static_cast<BYTE>(1);

  //Send the response
  try
  {
    m_pSocksRequest->m_ClientSocket.SendWithRetry(&userResponse, sizeof(userResponse), m_pSettings->m_dwWriteTimeout);
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
  catch(CWSocketException* pEx)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoUserNameAuthentication, Failed to send Socks5 user authentication packet, error:%d, client:%s"), pEx->m_nError, m_sLanAddress.GetString());
    m_pError->OnError(sError);

    pEx->Delete();

    return false;
  }
#else
  catch(CWSocketException& e)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoUserNameAuthentication, Failed to send Socks5 user authentication packet, error:%d, client:%s"), e.m_nError, m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  if (!bAllowUser)
  {
    //Report the error
    CString sError;
    sError.Format(_T("CSocksClient::DoUserNameAuthentication, Not allowing Socks5 user authentication from , User:%s, client:%s"), userDetails.sUserName.GetString(), m_sLanAddress.GetString());
    m_pError->OnError(sError);

    return false;
  }

  return true;
}

void CSocksClient::HandleClient()
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);
  ASSERT(m_pError != nullptr);

  //Form a string version of the lanaddress (helpful for logging)
  if (m_pSocksRequest->m_ClientAddress.si_family == AF_INET)
  {
    m_sLanAddress.Format(_T("%d.%d.%d.%d"), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_addr.S_un.S_un_b.s_b1), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_addr.S_un.S_un_b.s_b2), 
                         static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_addr.S_un.S_un_b.s_b3), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_addr.S_un.S_un_b.s_b4)); 
  }
  else
  {
    ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET6);
    m_sLanAddress.Format(_T("%04X.%04X.%04X.%04X.%04X.%04X.%04X.%04X"), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_addr.u.Word[0]), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_addr.u.Word[1]), 
                         static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_addr.u.Word[2]), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_addr.u.Word[3]), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_addr.u.Word[4]),
                         static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_addr.u.Word[5]), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_addr.u.Word[6]), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_addr.u.Word[7]));
  }

  if (!SetupInboundDetails())
    return;

  //Pick one of the connection methods the client provided
  if (m_ConnectionRequest.m_bSocks5Request)
  {
    if (!DoSocks5Negotiation())
      return;
  }
  else
  {
    //Drop the connection if we are not supporting Socks4
    if (!m_pSettings->m_bSupportSocks4)
    {
      //Report the error
      CString sError;
      sError.Format(_T("CSocksClient::HandleClient, Dropping client connection as we are not configured to support Socks 4 protocol, client:%s"), m_sLanAddress.GetString());
      m_pError->OnError(sError);

      return;
    }
  }

  ///Now try and make the connection to the remote host
  CSocksSocket destinationSocket;
  bool bSocketSetupFailure = false;
  int nConnectionError = ERROR_SUCCESS;
  bool bConnected = SetupOutboundDetails(destinationSocket, bSocketSetupFailure, nConnectionError);

  //Send the response back to the client
  if (m_ConnectionRequest.m_bSocks5Request)
  {
    if (!bConnected)
    {
      if (bSocketSetupFailure)
        SendConnectionResponsev5(1, INVALID_SOCKET);
      else if (nConnectionError == WSAECONNREFUSED)
        SendConnectionResponsev5(5, INVALID_SOCKET);
      else if (nConnectionError == WSAENETUNREACH)
        SendConnectionResponsev5(3, INVALID_SOCKET);
      else
        SendConnectionResponsev5(4, INVALID_SOCKET);
      return;
    }
    else if (!SendConnectionResponsev5(0, destinationSocket))
      return;
  }
  else
  {
    if (!bConnected)
    {
      SendConnectionResponsev4(false);
      return;
    }
    else if (!SendConnectionResponsev4(true))
      return;
  }

  //Now do the relaying on the sockets
  if (AllowThisOutboundConnection())
  {
    Relay(destinationSocket);
    OnEndRelay(); 
  }
}

BOOL CSocksClient::Run(_In_ const CThreadPoolRequest& request)
{
  //Validate our parameters
  ASSERT(request.m_pData != nullptr);
  ASSERT(m_pSettings != nullptr);
  ASSERT(m_pError != nullptr);
  m_pSocksRequest = static_cast<CSocksThreadPoolRequest*>(request.m_pData);
  ASSERT(m_pSocksRequest != nullptr);

  //Hand off the work to the helper function
  HandleClient();

  //Reset the member variables
  m_ConnectionRequest = CSocksConnectionRequest();
  m_RequestDetails = CSocks5RequestDetails();

  //Tidy up the heap memory
  delete m_pSocksRequest;
  m_pSocksRequest = nullptr;

  return TRUE;
}

void CSocksClient::SetRequestToStop()
{
  //Let the base class do its thing
  __super::SetRequestToStop();

  //Set the event which signals that we want this worker thread to exit
  m_StopEvent.SetEvent();
}

bool CSocksClient::AllowThisOutboundConnection()
{
#ifdef _DEBUG
  //Validate our parameters
  ASSERT(m_pSocksRequest != nullptr);

  //The default is to just TRACE the details
  if (m_ConnectionRequest.m_bSocks5Request)
  {
    if (m_pSocksRequest->m_ClientAddress.si_family == AF_INET)
    {
      TRACE(_T("Begining SOCKS5: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    }
    else
    {
      ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET6);
      TRACE(_T("Begining SOCKS5: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    }
  }
  else
  {
    ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET);
    TRACE(_T("Begining SOCKS4: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
  }
#endif //#ifdef _DEBUG

  return true;
}

void CSocksClient::OnEndRelay()
{
#ifdef _DEBUG
  //Validate our parameters
  ASSERT(m_pSocksRequest != nullptr);

  //The default is to just TRACE the details
  if (m_ConnectionRequest.m_bSocks5Request)
  {
    if (m_pSocksRequest->m_ClientAddress.si_family == AF_INET)
    {
      TRACE(_T("Ending SOCKS5: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    }
    else
    {
      ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET6);
      TRACE(_T("Ending SOCKS5: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    }
  }
  else
  {
    ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET);
    TRACE(_T("Ending SOCKS4: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
  }
#endif //#ifdef _DEBUG
}

void CSocksClient::Relay(_In_ CSocksSocket& destinationSocket)
{
  //Validate our parameters
  ASSERT(m_pSettings != nullptr);

  //Setup the handles to wait on
  HANDLE hWaitHandles[3];
  hWaitHandles[0] = m_DataEvent1;
  hWaitHandles[1] = m_DataEvent2;
  hWaitHandles[2] = m_StopEvent;

  while (TRUE)
  {
    //Wait for something interesting to happen
    DWORD dwWait = WaitForMultipleObjects(3, hWaitHandles, FALSE, INFINITE);
    int nSignaledHandle = dwWait - WAIT_OBJECT_0;

    //Work out what the return value from WFMO means!
    if (nSignaledHandle == 0)
    {
      //Work out what the event meant
      WSANETWORKEVENTS networkEvents;
      if (WSAEnumNetworkEvents(m_pSocksRequest->m_ClientSocket, m_DataEvent1, &networkEvents) == SOCKET_ERROR)
      {
        TRACE(_T("CSocksClient::Relay, Failed in call to WSAEnumNetworkEvents for LAN socket, error:%d"), WSAGetLastError());
        return;
      }
      if (networkEvents.lNetworkEvents & FD_CLOSE)
      {
        TRACE(_T("CSocksClient::Relay, returning because of FD_CLOSE event on LAN socket\n"));
        return;
      }

      //Read data from the LAN socket
      BYTE buffer[1024];
      int nData = 0;
      try
      {
        if (m_pSocksRequest->m_ClientSocket.IsReadible(0))
        {
          nData = m_pSocksRequest->m_ClientSocket.Receive(buffer, sizeof(buffer));
          if (nData)
          {
            //Write the data to the inet socket
            destinationSocket.SendWithRetry(buffer, nData, m_pSettings->m_dwWriteTimeout);
          }
          else
          {
            TRACE(_T("CSocksClient::Relay, stopping because of graceful disconnect on lan socket\n"));
            return;
          }
        }
      }
    #ifdef CWSOCKET_MFC_EXTENSIONS
      catch(CWSocketException* pEx)
      {
        TRACE(_T("CSocksClient::Relay, An error occurred relaying data from the lan socket to the internet socket, error:%d\n"), pEx->m_nError);
        pEx->Delete();
        return;
      }
    #else
      catch(CWSocketException& e)
      {
        UNREFERENCED_PARAMETER(e);
        TRACE(_T("CSocksClient::Relay, An error occurred relaying data from the lan socket to the internet socket, error:%d\n"), e.m_nError);
        return;
      }
    #endif //#ifdef CWSOCKET_MFC_EXTENSIONS
    }
    else if (nSignaledHandle == 1)
    {
      //Work out what the event meant
      WSANETWORKEVENTS networkEvents;
      if (WSAEnumNetworkEvents(destinationSocket, m_DataEvent2, &networkEvents) == SOCKET_ERROR)
      {
        TRACE(_T("CSocksClient::Relay, Failed in call to WSAEnumNetworkEvents for internet socket, error:%d"), WSAGetLastError());
        return;
      }
      if (networkEvents.lNetworkEvents & FD_CLOSE)
      {
        TRACE(_T("CSocksClient::Relay, returning because of FD_CLOSE event on internet socket\n"));
        return;
      }

      //Read data from the internet socket
      BYTE buffer[1024];
      int nData = 0;
      try
      {
        if (destinationSocket.IsReadible(0))
        {
          nData = destinationSocket.Receive(buffer, sizeof(buffer));
          if (nData)
          {
            //Write the data to the inet socket
            m_pSocksRequest->m_ClientSocket.SendWithRetry(buffer, nData, m_pSettings->m_dwWriteTimeout);
          }
          else
          {
            TRACE(_T("CSocksClient::Relay, stopping because of graceful disconnect on internet socket\n"));
            return;
          }
        }
      }
    #ifdef CWSOCKET_MFC_EXTENSIONS
      catch (CWSocketException* pEx)
      {
        TRACE(_T("CSocksClient::Relay, An error occurred reading data from the internet socket to the lan socket, error:%d\n"), pEx->m_nError);
        pEx->Delete();
        return;
      }
    #else
      catch(CWSocketException& e)
      {
        UNREFERENCED_PARAMETER(e);
        TRACE(_T("CSocksClient::Relay, An error occurred reading data from the internet socket to the lan socket, error:%d\n"), e.m_nError);
        return;
      }
    #endif //#ifdef CWSOCKET_MFC_EXTENSIONS
    }
    else
    {
      TRACE(_T("CSocksClient::Relay, thread being asked to exit\n"));
      return;
    }
  }
}
