/*
Module : SocksClient.h
Purpose: Defines the interface for the CSocksClient classes

Copyright (c) 2003 - 2017 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////////// Macros / Defines //////////////////////////////

#pragma once

#ifndef __SOCKSCLIENT_H__
#define __SOCKSCLIENT_H__


/////////////////////////////// Includes //////////////////////////////////////

#include "SocksSocket.h"
#include "ThrdPool.h" //If the code fails to compile because of this missing include then you need to download 
                      //the CWaitableTimer class from my web site and copy the relevant files to the SocksSvc directory


/////////////////////////////// Classes ///////////////////////////////////////

//Simply COM style error interface
class ISocksSvcError
{
public:
  virtual void OnError(_In_ const CString& sError) = 0;
};

//The settings which the socks server uses in the call to CSocksServer::Start
class CSocksServerSettings
{
public:
//Constructors / Destructors
  CSocksServerSettings() : m_nPort(1080),
                           m_dwReadTimeout(3000),
                           m_nThreadPoolSize(20),
                           m_dwWriteTimeout(3000),
                           m_bEnableThreadLifetime(false),
                           m_dwThreadLifetime(120),
                           m_bSupportSocks4(true),
                           m_bSupportSocks5(true),
                           m_bSupportAnonymous(true),
                           m_bSupportUserNamePassword(true),
                           m_bIPv6(false)
  {
  }

//Member variables
  unsigned short  m_nPort;                    //The port on which to listen for Socks request
  CString         m_sListenBindAddress;       //The IP address to bind to for the listening socket (empty implies no binding)
  CString         m_sConnectBindAddress;      //The IP address to bind to for the connecting socket (empty implies no binding)
  DWORD           m_dwReadTimeout;            //Timeout in ms to wait for sockets to become readable
  int             m_nThreadPoolSize;          //The size of the thread pool to use
  DWORD           m_dwWriteTimeout;           //Timeout in ms to wait for sockets to become writable
  bool            m_bEnableThreadLifetime;    //true if threads are to be recycled in the thread pool
  DWORD           m_dwThreadLifetime;         //The lifetime for threads (in Minutes) if recycling is enabled
  bool            m_bSupportSocks4;           //Should we support Sock4 
  bool            m_bSupportSocks5;           //Should we support Socks5
  bool            m_bSupportAnonymous;        //Should we support anonymous authentication
  bool            m_bSupportUserNamePassword; //Should we support username / password authentication
  CString         m_sUsername;                //The one and only valid username which is allowed if UserName authentication is in effect 
  CString         m_sPassword;                //The one and only valid password which is allowed if UserName authentication is in effect 
  bool            m_bIPv6;                    //Should the server listen on an IPv6 address
};

//Class which represents the thread pool request
class CSocksThreadPoolRequest
{
public:
  SOCKADDR_INET m_ClientAddress;
  CSocksSocket  m_ClientSocket;
};

//Class which handles the Socks client connection
class CSocksClient : public CThreadPoolClient
{
public:
//Constructors / Destructors
  CSocksClient();

//Methods
  //void SetServer(CSocksServer* pServer);
  virtual BOOL Run(_In_ const CThreadPoolRequest& request);
  virtual BOOL InitInstance();

//Member variables
  CSocksServerSettings*    m_pSettings;         //The settings to use
  ISocksSvcError*          m_pError;            //The error callback interface
 
protected:
//Methods
  virtual void SetRequestToStop();
  virtual bool AllowThisIncomingConnection();
  virtual bool SetupInboundDetails();
  virtual bool SetupOutboundDetails(_In_ CSocksSocket& destinationSocket, _Out_ bool& bSocketSetupFailure, _Out_ int& nConnectionError);
  virtual bool AllowUser(_In_ const CSocksUsernameAuthenticationDetails& userDetails);
  virtual bool AllowThisOutboundConnection();
  virtual bool SendConnectionResponsev5(_In_ BYTE byCMD, _In_ SOCKET hDestinationSocket);
  virtual bool SendConnectionResponsev4(_In_ bool bConnected);
  virtual void OnEndRelay();
  virtual void HandleClient();
  virtual void Relay(_In_ CSocksSocket& destinationSocket);
  virtual bool DoUserNameAuthentication();
  virtual bool DoSocks5Negotiation();

//Member variables
  CEvent                   m_StopEvent;         //Signaled when the server wants this client to exit
  CWaitableTimer           m_DataTimer;         //Used to implement read timeouts
  CEvent                   m_DataEvent1;        //Signaled when data is available on the incoming socket
  CEvent                   m_DataEvent2;        //Signaled when data is available on the connected socket
  CString                  m_sLanAddress;       //The IP address of the incoming client
  CString                  m_sRemoteAddress;    //IP address or host name of the connection to be made
  WORD                     m_nRemotePort;       //remote port number of the connection to be made
  CSocksThreadPoolRequest* m_pSocksRequest;     //Pointer to the actual thread pool request
  CSocksConnectionRequest  m_ConnectionRequest; //The initial connection details
  CSocks5RequestDetails    m_RequestDetails;    //The connection request details
};

#endif //#ifndef __SOCKSCLIENT_H__
